@@ -2,7 +2,7 @@
 # -----------------------------------------------------------------------------
 # config.py - config file reader
 # -----------------------------------------------------------------------------
-# $Id: config.py 2627 2007-04-15 20:52:39Z tack $
+# $Id: config.py 2950 2008-01-07 21:31:11Z dmeyer $
 #
 # -----------------------------------------------------------------------------
 # Copyright (C) 2006 Dirk Meyer, Jason Tackaberry
@@ -28,8 +28,8 @@
 #
 # -----------------------------------------------------------------------------
 
-__all__ = [ "Var", "Group", "Dict", "List", "Config", "set_default",
-            "get_description" ]
+__all__ = [ 'Var', 'Group', 'Dict', 'List', 'Config', 'set_default',
+            'get_description', 'get_config' ]
 
 # Python imports
 import os
@@ -38,11 +38,12 @@ import copy
 import logging
 import stat
 import md5
+import textwrap
 from new import classobj
 
 # kaa.base modules
 from strutils import str_to_unicode, unicode_to_str, get_encoding
-from kaa.notifier import Callback, WeakCallback, WeakTimer, WeakOneShotTimer
+from kaa import Callback, WeakCallback, WeakTimer, WeakOneShotTimer
 from kaa.inotify import INotify
 
 # get logging object
@@ -99,7 +100,7 @@ class Base(object):
         """
         Returns a hash of the config item.
         """
-        return md5.new(repr(self._name) + repr(self._desc) + repr(self._default)).hexdigest()
+        return md5.new(repr(self._name) + repr(self._desc) + repr(self._default) + repr(self._value)).hexdigest()
 
 
     def copy(self):
@@ -229,16 +230,35 @@ class Var(Base):
         Convert object into a string to write into a config file.
         """
         # create description
-        desc = newline = comment = ''
+        desc = comment = ''
         if print_desc:
             if self._desc:
-                desc = '# %s\n' % unicode_to_str(self._desc).replace('\n', '\n# ')
-            newline = '\n'
+                desc = '# | %s\n' % unicode_to_str(self._desc).replace('\n', '\n# | ')
+                if isinstance(self._type, (tuple, list)):
+                    # Show list of allowed values for this tuple variable.
+                    if type(self._type[0]) == type(self._type[-1]) == int and \
+                       self._type == range(self._type[0], self._type[-1]):
+                        # Type is a range
+                        allowed = '%d-%d' % (self._type[0], self._type[-1])
+                    else:
+                        allowed = ', '.join([ str(x) for x in self._type ])
+                    allowed = textwrap.wrap(allowed, 78, 
+                                          initial_indent = '# | Allowed values: ',
+                                          subsequent_indent = '# |' + 17 * ' ')
+                    desc += '\n'.join(allowed) + '\n'
+
+                if self._value != self._default:
+                    # User value is different than default, so include default
+                    # in comments for reference.
+                    desc += '# | Default: ' + str(self._default) + '\n'
+
         if self._value == self._default:
+            # Value is set to default, so comment it out in config file.
             comment = '# '
+
         value = unicode_to_str(self._value)
         prefix += self._name
-        return '%s%s%s = %s%s' % (desc, comment, prefix, value, newline)
+        return '%s%s%s = %s' % (desc, comment, prefix, value)
 
 
     def _cfg_set(self, value, default=False):
@@ -252,7 +272,8 @@ class Var(Base):
                 # This could crash, but that is ok
                 value = self._type[0].__class__(value)
             if not value in self._type:
-                raise AttributeError('Variable must be one of %s' % str(self._type))
+                allowed = [ str(x) for x in self._type ]
+                raise AttributeError('Variable must be one of %s' % ', '.join(allowed))
         elif not isinstance(value, self._type):
             if self._type == str:
                 value = unicode_to_str(value)
@@ -313,6 +334,13 @@ class Group(Base):
         self._vars.append(name)
 
 
+    def get_variables(self):
+        """
+        Returns a list of variables for this group.
+        """
+        return self._vars
+
+
     def _hash(self):
         """
         Returns a hash of the config item.
@@ -328,29 +356,57 @@ class Group(Base):
         Convert object into a string to write into a config file.
         """
         ret  = []
-        desc = unicode_to_str(self._desc).replace('\n', '\n# ')
+        desc = unicode_to_str(self._desc.strip('\n')).replace('\n', '\n# | ')
+        is_anonymous = self._name.endswith(']')
+
+        if print_desc and self._name and not is_anonymous:
+            sections = [ x.capitalize() for x in prefix.rstrip('.').split('.') + [self._name] ]
+            breadcrumb = ' > '.join(filter(len, sections))
+            ret.append('#\n# Begin Group: %s\n#' % breadcrumb)
+
         if self._name:
-            # add self._name to prefix and add a '.'
-            prefix = prefix + self._name + '.'
+            prefix = '%s%s.' % (prefix, self._name)
         print_var_desc = print_desc
-        if prefix and not prefix.endswith('].') and print_desc:
-            if not desc:
-                desc = 'group %s settings' % prefix[:-1]
-            ret.append('#\n# %s\n#\n' % desc)
-            if not self._desc_type == 'default':
+
+        if prefix and desc and not is_anonymous and print_desc:
+            ret.append('# | %s\n#' % desc)
+            if self._desc_type != 'default':
                 print_var_desc = False
 
+        # Iterate over group vars and fetch their cfg strings, also
+        # deterining if we need to space them (by separating with a
+        # blank commented line), which we do if
+        var_strings = []
+        space_vars = False
+        n_nongroup = 0
         for name in self._vars:
             var = self._dict[name]
-            if not isinstance(var, Var) or var._desc:
-                break
-        else:
-            print_var_desc = False
-        for name in self._vars:
-            var = self._dict[name]
-            ret.append(var._cfg_string(prefix, print_var_desc))
-        if print_desc and not print_var_desc:
-            ret.append('')
+            var_is_group = isinstance(var, (Group, Dict))
+            cfgstr = var._cfg_string(prefix, print_var_desc)
+            if not var_is_group:
+                n_nongroup += 1
+                space_vars = space_vars or '\n' in cfgstr
+            var_strings.append((cfgstr, var_is_group))
+
+        for (cfgstr, var_is_group) in var_strings:
+            if var_is_group and (not ret or not ret[-1].endswith('\n')):
+                # Config item is a group or list, space it down with a blank
+                # line for readability.
+                ret.append('')
+            ret.append(cfgstr)
+            if not var_is_group and space_vars and n_nongroup > 1:
+                # We need to space variables (see above), so add the empty
+                # commented line.
+                ret.append('#')
+
+        if print_desc and self._name and not is_anonymous:
+            if n_nongroup != len(self._vars) and (not ret or not ret[-1].endswith('\n')):
+                # One of our variables is a group/dict, so add another
+                # empty line to separate the stanza.
+                ret.append('\n#')
+            elif not space_vars or n_nongroup <= 1:
+                ret.append('#')
+            ret.append('# End Group: %s\n#\n' % breadcrumb)
         return '\n'.join(ret)
 
 
@@ -454,16 +510,40 @@ class Dict(Base):
         Convert object into a string to write into a config file.
         """
         ret = []
+        if print_desc:
+            sections = [ x.capitalize() for x in prefix.rstrip('.').split('.') + [self._name] ]
+            breadcrumb = ' > '.join(filter(len, sections))
+            ret.append('#\n# Begin %s: %s\n#' % (self.__class__.__name__, breadcrumb))
+
         prefix = prefix + self._name
-        if (type(self._schema) == Var and print_desc) or not self.keys():
+
+        if print_desc: #(type(self._schema) == Var and print_desc) or not self.keys():
             # TODO: more detailed comments, show full spec of var and some examples.
-            d = unicode_to_str(self._desc).replace('\n', '\n# ')
-            ret.append('#\n# %s\n# %s\n#\n' % (prefix, d))
-            print_desc = False
+            ret.append('# | %s' % prefix)
+            if self._desc:
+                desc = unicode_to_str(self._desc).replace('\n', '\n# | ')
+                ret.append('# |\n# | %s' % desc)
+
+        var_strings = []
+        space_vars = False
         for key in self.keys():
-            ret.append(self._dict[key]._cfg_string(prefix, print_desc))
-        if not print_desc:
-            ret.append('')
+            cfgstr = self._dict[key]._cfg_string(prefix, False)
+            var_strings.append(cfgstr)
+            if '\n' in cfgstr:
+                space_vars = True
+
+        for cfgstr in var_strings:
+            if '# Begin' in cfgstr and (not ret or not ret[-1].endswith('\n')):
+                # Config item is a group or list, space it down with a blank
+                # line for readability.
+                ret.append('')
+            ret.append(cfgstr)
+            if space_vars:
+                # Separate multi-line subgroups with newline. 
+                ret.append('')
+
+        if print_desc:
+            ret.append('#\n# End %s: %s\n#' % (self.__class__.__name__, breadcrumb))
         return '\n'.join(ret)
 
 
@@ -566,11 +646,12 @@ class Config(Group):
     """
     A config object. This is a group with functions to load and save a file.
     """
-    def __init__(self, schema, desc=u'', name=''):
+    def __init__(self, schema, desc=u'', name='', module = None):
         Group.__init__(self, schema, desc, name)
         self._filename = None
         self._bad_lines = []
         self._loaded_hash = None
+        self._module = module
 
         # Whether or not to autosave config file when options have changed
         self._autosave = None
@@ -604,17 +685,10 @@ class Config(Group):
         return copy
 
 
-    def save(self, filename = None, force = False):
+    def save(self, filename = None):
         """
         Save file. If filename is not given use filename from last load.
-        If force is True, the config file will be saved even if there were
-        no schema changes from the previously loaded file.
         """
-        hash = self._hash()
-        if not force and hash == self._loaded_hash:
-            # Hash hasn't changed, so we don't need to write the file.
-            return
-
         if not filename:
             if not self._filename:
                 raise ValueError, "Filename not specified and no default filename set."
@@ -626,9 +700,12 @@ class Config(Group):
 
         self._autosave_timer.stop()
         f = open(filename, 'w')
-        f.write('# -*- coding: %s -*-\n' % get_encoding().lower() + \
-                '# -*- hash: %s -*-\n' % self._hash() + \
-                '# *************************************************************\n' + \
+        encoding = get_encoding().lower().replace('iso8859', 'iso-8859')
+        f.write('# -*- coding: %s -*-\n' % encoding + \
+                '# -*- hash: %s -*-\n' % self._hash())
+        if self._module:
+            f.write('# -*- module: %s -*-\n' % self._module)
+        f.write('# *************************************************************\n' + \
                 '# WARNING: This file is auto-generated.  You are free to edit\n' + \
                 '# this file to change config values, but any other changes\n' + \
                 # FIXME: custom comments lost, would be nice if they were kept.  Might
@@ -749,7 +826,10 @@ class Config(Group):
         self._watch_mtime = os.stat(filename)[stat.ST_MTIME]
         if create:
             # Write config file if schema is different.
-            self.save(filename)
+            hash = self._hash()
+            if hash != self._loaded_hash:
+                self.save(filename)
+
         return len(self._bad_lines) == 0
 
 
@@ -847,8 +927,16 @@ class Config(Group):
 
 
     def _file_changed(self, mask, path):
-        if mask & (INotify.MODIFY | INotify.ATTRIB):
+        if mask & INotify.MODIFY:
+            # Config file changed.  Attach a monitor so we can keep track of
+            # any values that actually changed.
+            changed_names = []
+            cb = Callback(lambda *args: changed_names.append(args[0]))
+            self.add_monitor(cb)
             self.load()
+            log.info('Config file %s modified; %d settings changed.' % (self._filename, len(changed_names)))
+            log.debug('What changed: %s', ', '.join(changed_names) or 'nothing')
+            self.remove_monitor(cb)
         elif mask & (INotify.IGNORED | INotify.MOVE_SELF):
             # File may have been replaced, check mtime now.
             WeakOneShotTimer(self._check_file_changed).start(0.1)
@@ -858,13 +946,72 @@ class Config(Group):
 
 def set_default(var, value):
     """
-    Set default value for the given config variable.
+    Set default value for the given config variable (proxy).
     """
-    var._item._cfg_set(value, default = True)
+    if isinstance(var, VarProxy):
+        var._item._cfg_set(value, default = True)
 
 
+def get_default(var):
+    if isinstance(var, VarProxy):
+        return var._item._default
+    
+
 def get_description(var):
     """
-    Get the description for the given config variable.
+    Get the description for the given config variable or group.
     """
-    return var._item._desc
+    if isinstance(var, (Group, Dict)):
+        return var._desc
+    elif isinstance(var, VarProxy):
+        return var._item._desc
+
+
+def get_type(var):
+    """
+    Returns the type of the given config variable.
+    """
+    if isinstance(var, VarProxy):
+        return var._item._type
+    return type(var)
+    
+
+def get_config(filename, module = None):
+    """
+    Returns a Config object representing the config file provided in
+    'filenane'.  If module is None, the specified config file must have the
+    module specified (in the "-*- module: ... -*-" metadata), otherwise the
+    supplied module (string) is used.  The module must be importable.  
+    
+    If the config module cannot be determined and one is not specified,
+    will raise ValueError.  If import fails, will raise ImportError.
+    Otherwise will return the Config object.
+    """
+    filename = os.path.expanduser(filename)
+
+    if not module:
+        # No module specified, check the config file.
+        metadata = file(filename).read(256)
+        m = re.search(r'^# -\*- module: (\S+) -\*-$', metadata, re.M)
+        if m:
+            module = m.group(1)
+        else:
+            raise ValueError, 'No module specified in config file'
+
+    components = module.split('.')
+    attr = components.pop()
+    module = '.'.join(components)
+    try:
+        exec('import %s as module' % module)
+    except Exception, e:
+        log.exception('config loader')
+        raise ImportError, 'Could not import config module %s' % module
+
+    config = getattr(module, attr)
+    if config._filename and os.path.realpath(config._filename) != os.path.realpath(filename):
+        # Existing config object represents a different config file,
+        # so must copy.
+        config = config.copy()
+
+    config.load(filename)
+    return config
