@@ -2,7 +2,7 @@
 # -----------------------------------------------------------------------------
 # core.py - distribution core functions for kaa packages
 # -----------------------------------------------------------------------------
-# $Id: core.py 2605 2007-03-30 15:03:33Z dmeyer $
+# $Id: core.py 3206 2008-03-20 14:16:48Z dmeyer $
 #
 # -----------------------------------------------------------------------------
 # Copyright (C) 2006 Dirk Meyer, Jason Tackaberry
@@ -80,9 +80,9 @@ def compile(includes, code='', args=''):
     return False
 
 
-def check_library(name, *args):
+def check_library(name, *args, **kwargs):
     lib = Library(name)
-    if len(args) < 2:
+    if len(args) < 2 and not kwargs:
         minver = args[0]
         print 'checking for', name, '>=', minver, '...',
         sys.__stdout__.flush()
@@ -91,15 +91,21 @@ def check_library(name, *args):
             print version
         elif version:
             print 'no (%s)' % version
+            return
         else:
             print 'no'
+            return
     else:
+        for var in ('include_dirs', 'library_dirs', 'libraries'):
+            if var in kwargs:
+                setattr(lib, var, kwargs.pop(var))
         print 'checking for', name, '...',
         sys.__stdout__.flush()
-        if lib.compile(*args):
+        if lib.compile(*args, **kwargs):
             print 'ok'
         else:
             print 'no'
+            return
     _libraries[name] = lib
     return lib
 
@@ -203,6 +209,8 @@ class Library(object):
             for dir in lib.include_dirs:
                 ext_args.append("-I%s" % dir)
 
+        # FIXME: we do not check if the lib is installed, we only
+        # check the header files.
         if compile(includes, code, args + ' %s' % ' '.join(ext_args)):
             self.valid = True
             return True
@@ -330,6 +338,14 @@ class Extension(object):
         return compile(includes, code, args)
 
 
+    def has_python_h(self):
+        """
+        Return True if Python.h is found in the system. If not compiling
+        the extention will fail in most cases.
+        """
+        return os.path.exists(os.path.join(distutils.sysconfig.get_python_inc(), 'Python.h'))
+
+        
     def convert(self):
         """
         Convert Extension into a distutils.core.Extension.
@@ -396,14 +412,18 @@ def setup(**kwargs):
         """
         Helper function to create 'packages' and 'package_dir'.
         """
-        if not '__init__.py' in files:
-            return
-        python_dirname = prefix + dirname[3:].replace('/', '.')
-        # Anything under module/src/extensions/foo gets translated to 
-        # kaa.module.foo.
-        python_dirname = python_dirname.replace(".extensions.", ".")
-        kwargs['package_dir'][python_dirname] = dirname
-        kwargs['packages'].append(python_dirname)
+        for key, value in kwargs.get('plugins', {}).items():
+            if dirname.startswith(value):
+                python_dirname = key + dirname[len(value):].replace('/', '.')
+                break
+        else:
+            python_dirname = prefix + dirname[3:].replace('/', '.')
+            # Anything under module/src/extensions/foo gets translated to 
+            # kaa.module.foo.
+            python_dirname = python_dirname.replace(".extensions.", ".")
+        if '__init__.py' in files or python_dirname.endswith('plugins'):
+            kwargs['package_dir'][python_dirname] = dirname
+            kwargs['packages'].append(python_dirname)
 
 
     if 'module' not in kwargs:
@@ -420,6 +440,9 @@ def setup(**kwargs):
     else:
         os.path.walk('src', _find_packages, (kwargs, 'kaa.' + kwargs['module']))
 
+    if 'plugins' in kwargs:
+        del kwargs['plugins']
+        
     # convert Extensions
     if kwargs.get('ext_modules'):
         kaa_ext_modules = kwargs['ext_modules']
@@ -440,7 +463,7 @@ def setup(**kwargs):
 
     # check version.py information
     write_version = False
-    if 'version' in kwargs and not kwargs['module'] == 'base':
+    if 'version' in kwargs:
         write_version = True
         # check if a version.py is there
         if os.path.isfile('src/version.py'):
@@ -452,7 +475,7 @@ def setup(**kwargs):
                         write_version = False
                     break
             f.close()
-
+           
     if len(sys.argv) > 1 and sys.argv[1] in ('bdist_rpm', 'sdist') and \
            os.path.isfile('ChangeLog.in'):
         # FIXME: find a better way to detect if we need to create a
@@ -463,13 +486,13 @@ def setup(**kwargs):
     # delete 'module' information, not used by distutils.setup
     del kwargs['module']
 
+    # Write a version.py and add it to the list of files to
+    # be installed.
     if write_version:
-        # Write a version.py and add it to the list of files to
-        # be installed.
         f = open('src/version.py', 'w')
         f.write('# version information for %s\n' % kwargs['name'])
         f.write('# autogenerated by kaa.distribution\n\n')
-        f.write('from kaa.version import Version\n')
+        f.write('from kaa.distribution.version import Version\n')
         f.write('VERSION = Version(\'%s\')\n' % kwargs['version'])
         f.close()
 
