@@ -2,7 +2,7 @@
 # -----------------------------------------------------------------------------
 # popen.py - process control using notifier
 # -----------------------------------------------------------------------------
-# $Id: popen.py 2606 2007-03-30 20:28:32Z dmeyer $
+# $Id: popen.py 3104 2008-02-15 14:33:33Z tack $
 #
 # This module defines a process class similar to the once defined in popen2
 # except that this class is aware of the notifier loop.
@@ -15,12 +15,12 @@
 # By inherting from the class you can also override the functions stdout_cb
 # and stderr_cb to process stdout and stderr line by line.
 #
-# The killall function of this class can be called at the end of the programm
+# The kill_all function of this class can be called at the end of the programm
 # to stop all running processes.
 #
 # -----------------------------------------------------------------------------
 # kaa.notifier - Mainloop and callbacks
-# Copyright (C) 2005, 2006 Dirk Meyer, Jason Tackaberry, et al.
+# Copyright (C) 2005-2008 Dirk Meyer, Jason Tackaberry, et al.
 #
 # First Version: Dirk Meyer <dmeyer@tzi.de>
 # Maintainer:    Dirk Meyer <dmeyer@tzi.de>
@@ -44,7 +44,7 @@
 #
 # -----------------------------------------------------------------------------
 
-__all__ = [ 'Process', 'stop_all_processes', 'kill_all_processes']
+__all__ = [ 'Process', 'proclist' ]
 
 # python imports
 import os
@@ -56,13 +56,16 @@ import logging
 
 # notifier imports
 import nf_wrapper as notifier
-from callback import Signal, Callback
+from callback import Callback
+from signals import Signal
 from thread import MainThreadCallback, is_mainthread
 from async import InProgress
+from sockets import IOMonitor, IO_WRITE
 
 # get logging object
 log = logging.getLogger('notifier')
 
+# FIXME: rewrite :)
 
 class Process(object):
     """
@@ -80,6 +83,8 @@ class Process(object):
         self.signals = {
             "stderr": Signal(),
             "stdout": Signal(),
+            "raw-stderr": Signal(),
+            "raw-stdout": Signal(),
             "completed": Signal(),
         }
 
@@ -91,6 +96,9 @@ class Process(object):
         self.__kill_timer = None
         self.child = None
         self.in_progress = None
+        self._write_buffer = []
+        self._wmon = None
+        self._close_stdin = False
 
 
     def _normalize_cmd(self, cmd):
@@ -151,34 +159,96 @@ class Process(object):
 
         self.child = popen2.Popen3( cmd, True, 100 )
 
+        flags = fcntl.fcntl(self.child.tochild.fileno(), fcntl.F_GETFL)
+        fcntl.fcntl( self.child.tochild.fileno(), fcntl.F_SETFL, flags | os.O_NONBLOCK )
+        self._wmon = IOMonitor(self._handle_write)
+        if self._write_buffer:
+            self._wmon.register(self.child.tochild, IO_WRITE)
+
         log.info('running %s (pid=%s)' % ( self.binary, self.child.pid ) )
 
         # IO_Handler for stdout
         self.stdout = IO_Handler( 'stdout', self.child.fromchild,
-                                  self.signals["stdout"].emit, self._debugname )
+                                  self.signals["stdout"], self.signals['raw-stdout'], self._debugname )
         # IO_Handler for stderr
         self.stderr = IO_Handler( 'stderr', self.child.childerr,
-                                  self.signals["stderr"].emit, self._debugname )
+                                  self.signals["stderr"], self.signals['raw-stderr'], self._debugname )
 
         # add child to watcher
         if not is_mainthread():
-            MainThreadCallback(_watcher.append, self, self.__child_died )
+            MainThreadCallback(proclist.append)(self, self.__child_died)
         else:
-            _watcher.append( self, self.__child_died )
+            proclist.append( self, self.__child_died )
         self.in_progress = InProgress()
         return self.in_progress
 
 
-    def write( self, line ):
+    def get_pid(self):
         """
-        Write a string to the app.
+        Returns the pid of the child process if it has been spawned.  Otherwise
+        returns None
         """
+        if self.child:
+            return self.child.pid
+
+
+    def close_stdin(self):
+        """
+        Closes stdin either now, or once the write buffer has been flushed.
+        This might be necessary when, for example, we know we have nothing
+        more to send to the child, but the child is waiting for either more
+        data or a closed fd before outputting data we're interested in.
+        """
+        if not self.child:
+            return
+
+        if self._wmon and self._wmon.active():
+            self._close_stdin = True
+        else:
+            self.child.tochild.close()
+
+
+    def write(self, data):
+        """
+        Queue data for writing when the child is ready to receive it.
+        """
+        if self.child.tochild.closed:
+            raise ValueError("Can't write when stdin has been closed")
+
+        self._write_buffer.append(data)
+        if self.child and self._wmon and not self._wmon.active():
+            self._wmon.register(self.child.tochild, IO_WRITE)
+
+
+    def _handle_write(self):
+        if not self.child:
+            # The child process might have died before we had a chance to
+            # flush the write buffer.
+            del self._write_buffer[:]
+            return False
+
         try:
-            if self.child:
-                self.child.tochild.write(line)
-                self.child.tochild.flush()
-        except (IOError, ValueError):
-            pass
+            while self._write_buffer:
+                data = self._write_buffer[0]
+                sent = os.write(self.child.tochild.fileno(), data)
+                # If we got here, no exception was raised, so we can pop the
+                # data from the buffer.
+                self._write_buffer.pop(0)
+                if sent != len(data):
+                    # Not all data was sent, so push the remaining bytes back and
+                    # abort the loop.
+                    self._write_buffer.insert(0, data[sent:])
+                    break
+        except IOError, (errno, msg):
+            if errno == 11:
+                # Resource temporarily unavailable -- trying to write too
+                # much data.
+                return
+
+        if not self._write_buffer:
+            if self._close_stdin:
+                self.child.tochild.close()
+            return False
 
 
     def is_alive( self ):
@@ -302,10 +372,15 @@ class Process(object):
         return False
 
 
-    def __child_died( self, status ):
+    def __child_died( self, pid, status ):
         """
         Callback from watcher when the child died.
         """
+        if pid != self.get_pid():
+            # We received notification from the watcher concerning another
+            # process, so ignore.
+            return
+
         self.__dead = True
         self.stopping = False
         # close IO handler and kill timer
@@ -315,7 +390,7 @@ class Process(object):
         self.child = None
         if self.__kill_timer:
             notifier.timer_remove( self.__kill_timer )
-        self.in_progress.finished(status >> 8)
+        self.in_progress.finish(status >> 8)
         self.in_progress = None
         self.signals['completed'].emit(status >> 8)
 
@@ -325,12 +400,13 @@ class IO_Handler(object):
     """
     Reading data from socket (stdout or stderr)
     """
-    def __init__( self, name, fp, callback, logger = None):
+    def __init__( self, name, fp, signal, raw_signal, logger = None):
         self.name = name
         self.fp = fp
         flags = fcntl.fcntl(self.fp.fileno(), fcntl.F_GETFL)
         fcntl.fcntl( self.fp.fileno(), fcntl.F_SETFL, flags | os.O_NONBLOCK )
-        self.callback = callback
+        self.signal = signal
+        self.raw_signal = raw_signal
         self.logger = None
         self.saved = ''
         notifier.socket_add( fp, self._handle_input )
@@ -364,20 +440,23 @@ class IO_Handler(object):
         """
         Read as much as possible and close socket later.
         """
-        while self._handle_input( self.fp ):
+        while self._handle_input( self.fp, True ):
             pass
 
         
-    def _handle_input( self, socket ):
+    def _handle_input( self, socket, flushing = False ):
         """
         Handle data input from socket.
         """
         try:
-            data = self.fp.read( 10000 )
+            data = self.fp.read(512*1024)
         except IOError, (errno, msg):
-            if errno == 11:
+            if errno == 11 and not flushing:
                 # Resource temporarily unavailable; if we try to read on a
-                # non-blocking descriptor we'll get this message.
+                # non-blocking descriptor we'll get this message.  If we're
+                # being called from flush(), it could be because the child
+                # process is dead, in which case this errno will occur but
+                # we don't want to return True.
                 return True
             data = None
         except ValueError:
@@ -390,8 +469,19 @@ class IO_Handler(object):
             self.fp.close()
             if self.logger:
                 self.logger.close()
+            # FIXME: we know the child is dead.  Rather than wait for it
+            # to be reaped we can call __child_died in Process instance.
             return False
 
+        if self.raw_signal.count():
+            self.raw_signal.emit(data)
+
+        if self.signal.count() == 0:
+            # Nothing connected to the per-line handler, so need to parse.
+            # (And if we're dealing with binary data, it may be important
+            # _not to parse.)
+            return True
+
         data  = data.replace('\r', '\n')
         lines = data.split('\n')
 
@@ -403,7 +493,7 @@ class IO_Handler(object):
         # Combine saved data and first line, send to app
         if self.logger:
             self.logger.write( self.saved + lines[ 0 ] + '\n' )
-        self.callback( self.saved + lines[ 0 ] )
+        self.signal.emit( self.saved + lines[ 0 ] )
         self.saved = ''
 
         # There's one or more lines + possibly a partial line
@@ -417,7 +507,7 @@ class IO_Handler(object):
                     continue
                 if self.logger:
                     self.logger.write( line + '\n' )
-                self.callback( line )
+                self.signal.emit( line )
         else:
             # Send all lines to the app
             for line in lines[ 1 : ]:
@@ -425,7 +515,7 @@ class IO_Handler(object):
                     continue
                 if self.logger:
                     self.logger.write( line + '\n' )
-                self.callback( line )
+                self.signal.emit( line )
         return True
 
 
@@ -440,6 +530,7 @@ class Watcher(object):
         self.__processes[ proc ] = cb
         if not self.__timer:
             log.info('start process watching')
+            # FIXME: there is no reason to reap 20 times a second.
             self.__timer = notifier.timer_add(50, self.check)
 
 
@@ -448,25 +539,33 @@ class Watcher(object):
 
         # check all processes
         for p in self.__processes:
+
+            if isinstance( p.child, popen2.Popen3 ):
+                pid = p.child.pid
+            else:
+                pid = p.pid
+
             try:
-                if isinstance( p.child, popen2.Popen3 ):
-                    pid, status = os.waitpid( p.child.pid, os.WNOHANG )
-                else:
-                    pid, status = os.waitpid( p.pid, os.WNOHANG )
-            except OSError:
-                remove_proc.append( p )
+                pid, status = os.waitpid( pid, os.WNOHANG )
+            except OSError, (errno, err):
+                # FIXME: sometimes we get ECHILD ("No child processes", errno=10) even
+                # though we definitely have children.  This happens with pppd, for
+                # example.  Need to figure out why.
+                remove_proc.append( (p, pid, -1) )
                 continue
+
             if not pid:
                 continue
+
             log.info('Dead child: %s (%s)' % ( pid, status ))
             if status == -1:
                 log.error('error retrieving process information from %d' % p)
             elif os.WIFEXITED( status ) or os.WIFSIGNALED( status ) or \
                      os.WCOREDUMP( status ):
-                remove_proc.append( p )
+                remove_proc.append( (p, pid, status) )
 
         # remove dead processes
-        for p in remove_proc:
+        for p, pid, status in remove_proc:
             if p in self.__processes:
                 # call stopped callback
                 callback = self.__processes[p]
@@ -474,7 +573,7 @@ class Watcher(object):
                 # it, since it's possible the callback could call append
                 # again.
                 del self.__processes[p]
-                callback(status)
+                callback(pid, status)
 
         # check if this function needs to be called again
         if not self.__processes:
@@ -487,7 +586,7 @@ class Watcher(object):
         return True
 
 
-    def stopall( self ):
+    def stop_all( self ):
         if self.status != 'running':
             return
         # stop all childs without waiting
@@ -496,28 +595,14 @@ class Watcher(object):
         self.status = 'stopping'
 
 
-    def killall( self ):
+    def kill_all( self ):
         # prevent recursion
         if not self.status in ('running', 'stopping'):
             return
         # make sure every child is stopped
-        self.stopall()
+        self.stop_all()
         self.status = 'stopped'
 
-        # now wait until all childs are dead
-        while self.__processes:
-            self.check()
-            try:
-                notifier.step()
-            except ( KeyboardInterrupt, SystemExit ), e:
-                pass
-            except:
-                log.exception( 'Unhandled exception during killall' )
-
 
 # global watcher instance
-_watcher = Watcher()
-
-# global killall function
-stop_all_processes = _watcher.stopall
-kill_all_processes = _watcher.killall
+proclist = Watcher()
