@@ -1,10 +1,10 @@
-from kaa.notifier import execute_in_timer, Timer, OneShotTimer, loop, \
-     is_mainthread, execute_in_thread, execute_in_mainloop
+from kaa import timed, Timer, OneShotTimer, main, \
+     is_mainthread, threaded, MAINTHREAD, POLICY_ONCE, POLICY_RESTART
 
 
 class Foo(object):
 
-    @execute_in_timer(Timer, 0.1, 'once')
+    @timed(0.1, policy=POLICY_ONCE)
     def poll(self, x):
         if not x:
             return False
@@ -12,7 +12,7 @@ class Foo(object):
         print '1', y
         return True
 
-    @execute_in_timer(Timer, 0.1, 'once')
+    @timed(0.1, policy=POLICY_ONCE)
     def poll2(self, x):
         if not x:
             return False
@@ -20,17 +20,17 @@ class Foo(object):
         print '2', y
         return True
 
-    @execute_in_thread('name')
+    @threaded('name')
     def foo(self):
         print 'foo is mainthread:', is_mainthread()
         self.bar()
         
-    @execute_in_mainloop()
+    @threaded(MAINTHREAD)
     def bar(self):
         print 'bar is mainthread:', is_mainthread()
         
 
-@execute_in_timer(Timer, 0.1, 'override')
+@timed(0.1, policy=POLICY_RESTART)
 def poll(x):
     if not x:
         return False
@@ -39,7 +39,7 @@ def poll(x):
     return True
 
 
-@execute_in_timer(OneShotTimer, 0.1)
+@timed(0.1, OneShotTimer, policy=POLICY_RESTART)
 def bla(f, msg):
     print msg
     f.foo()
@@ -54,4 +54,4 @@ bla(f, 'test')
 bla(f, 'test2')
 OneShotTimer(poll, [20,21,22,23,24]).start(0.3)
 OneShotTimer(f.poll, [30,31,32,33,34]).start(0.3)
-loop()
+main.run()
